/********************************************************************** 
 techtree - Copyright (C) 1998 - David W. Pfitzner
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#include <math.h>

#include "techtree.h"
#include "metrics.h"

extern float dist_dy_factor;	/* in techtree.c */

/* node-node distance (direct, based on dx,dy) */
float
dist_node_node( tech_node_t *node1, tech_node_t *node2 )
{
    float dx, dy;
    
    if( dist_dy_factor == 0 ) {
	return fabs(node1->dx - node2->dx);
    } else {
	dx = node1->dx - node2->dx;
	dy = dist_dy_factor*(node1->dy - node2->dy);
	return sqrt(dx*dx + dy*dy);
    }
}

/* sums of lengths for immediate allow links of this node */
float
dist_node_allow( tech_tree_t *ttree, tech_node_t *node )
{
    tech_node_t *lnode;
    int k;
    float result = 0;

    for( k=0; k<node->n_allow; k++ ) {
	lnode = TNODE_P(ttree, node->i_allow[k]);
	result += dist_node_node( node, lnode );
    }
    return result;
}

/* ditto req links */
float
dist_node_req( tech_tree_t *ttree, tech_node_t *node )
{
    tech_node_t *lnode;
    int k;
    float result = 0;

    for( k=0; k<node->n_req; k++ ) {
	lnode = TNODE_P(ttree, node->i_req[k]);
	result += dist_node_node( node, lnode );
    }
    return result;
}

/* allow + req links */
float
dist_node_links( tech_tree_t *ttree, tech_node_t *node )
{
    return dist_node_allow(ttree,node)
	+ dist_node_req(ttree,node);
}

/* all links; do allow (+ req would be twice) */
float
dist_tree( tech_tree_t *ttree )
{
    tech_node_t *node;
    int i;
    float result = 0;

    for( i=0, node=ttree->nodes; i<ttree->n; i++, node++ ) {
	result += dist_node_allow(ttree,node);
    }
    return result;
}

/* number of crossings for "allow" links;
   the tree must have per_level[], level_start[] setup
   uses dx values
*/
int
cross_node_allow( tech_tree_t *ttree, tech_node_t *node )
{
    tech_node_t *a_node, *n1, *n2;
    int i, j1, j2, k1, k2, level, result;

    level = node->level;
    result = 0;
    for( i=0; i<ttree->per_level[level]; i++ ) {
	a_node = TNODE_P(ttree, ttree->i_sort[ttree->level_start[level]+i]);
	if( a_node == node ) {
	    continue;
	}
	k1 = (node->dx < a_node->dx);
	for( j1=0; j1<node->n_allow; j1++ ) {
	    n1 = TNODE_P(ttree, node->i_allow[j1]);
	    for( j2=0; j2<a_node->n_allow; j2++ ) {
		n2 = TNODE_P(ttree, a_node->i_allow[j2]);
		k2 = (n1->dx < n2->dx);
		result += ( (k1&&(!k2)) || ((!k1)&&k2) );
	    }
	}
    }
    return result;
}

/* same as above with allow->req */
int
cross_node_req( tech_tree_t *ttree, tech_node_t *node )
{
    tech_node_t *a_node, *n1, *n2;
    int i, j1, j2, k1, k2, level, result;

    level = node->level;
    result = 0;
    for( i=0; i<ttree->per_level[level]; i++ ) {
	a_node = TNODE_P(ttree, ttree->i_sort[ttree->level_start[level]+i]);
	if( a_node == node ) {
	    continue;
	}
	k1 = (node->dx < a_node->dx);
	for( j1=0; j1<node->n_req; j1++ ) {
	    n1 = TNODE_P(ttree, node->i_req[j1]);
	    for( j2=0; j2<a_node->n_req; j2++ ) {
		n2 = TNODE_P(ttree, a_node->i_req[j2]);
		k2 = (n1->dx < n2->dx);
		result += ( (k1&&(!k2)) || ((!k1)&&k2) );
	    }
	}
    }
    return result;
}

int
cross_node_links( tech_tree_t *ttree, tech_node_t *node )
{
    return cross_node_allow(ttree,node) +
	cross_node_req(ttree,node);
}

int
cross_tree( tech_tree_t *ttree )
{
    tech_node_t *node;
    int i, result;

    result = 0;
    for( i=0, node=ttree->nodes; i<ttree->n; i++, node++ ) {
	result += cross_node_allow(ttree,node);
    }
    return result/2;		/* count each intersection twice */
}

float
distx_node_allow( tech_tree_t *ttree, tech_node_t *node )
{
    tech_node_t *a_node, *n1, *n2;
    int i, j1, j2, k1, k2, level, nx_this;
    float result, dist_this;

    level = node->level;
    result = 0;
    for( j1=0; j1<node->n_allow; j1++ ) {
	n1 = TNODE_P(ttree, node->i_allow[j1]);
	dist_this = dist_node_node( node, n1 );
	nx_this = 1;
	for( i=0; i<ttree->per_level[level]; i++ ) {
	    a_node = TNODE_P(ttree, ttree->i_sort[ttree->level_start[level]+i]);
	    if( a_node == node ) {
		continue;
	    }
	    k1 = (node->dx < a_node->dx);
	    for( j2=0; j2<a_node->n_allow; j2++ ) {
		n2 = TNODE_P(ttree, a_node->i_allow[j2]);
		k2 = (n1->dx < n2->dx);
		nx_this += ( (k1&&(!k2)) || ((!k1)&&k2) );
	    }
	}
	result += dist_this * nx_this;
    }
    return result;
}

float
distx_node_req( tech_tree_t *ttree, tech_node_t *node )
{
    tech_node_t *a_node, *n1, *n2;
    int i, j1, j2, k1, k2, level, nx_this;
    float result, dist_this;

    level = node->level;
    result = 0;
    for( j1=0; j1<node->n_req; j1++ ) {
	n1 = TNODE_P(ttree, node->i_req[j1]);
	dist_this = dist_node_node( node, n1 );
	nx_this = 1;
	for( i=0; i<ttree->per_level[level]; i++ ) {
	    a_node = TNODE_P(ttree, ttree->i_sort[ttree->level_start[level]+i]);
	    if( a_node == node ) {
		continue;
	    }
	    k1 = (node->dx < a_node->dx);
	    for( j2=0; j2<a_node->n_req; j2++ ) {
		n2 = TNODE_P(ttree, a_node->i_req[j2]);
		k2 = (n1->dx < n2->dx);
		nx_this += ( (k1&&(!k2)) || ((!k1)&&k2) );
	    }
	}
	result += dist_this * nx_this;
    }
    return result;
}

float
distx_node_links( tech_tree_t *ttree, tech_node_t *node )
{
    return distx_node_allow(ttree,node) +
	distx_node_req(ttree,node);
}

float
distx_tree( tech_tree_t *ttree )
{
    tech_node_t *node;
    int i;
    float result;

    result = 0.0;
    for( i=0, node=ttree->nodes; i<ttree->n; i++, node++ ) {
	result += distx_node_allow(ttree,node);
    }
    return result;
}

void
calc_tree_metrics( tech_tree_t *ttree )
{
    ttree->metric_dist = dist_tree(ttree);
    ttree->metric_cross = cross_tree(ttree);
    ttree->metric_distx = distx_tree(ttree);
}

/* must have already called calc_tree_metrics */
void
print_tree_metrics( FILE *fp, tech_tree_t *ttree, int id, char *str )
{
    fprintf( fp, "%g %d %g %d # %s (D,X,DX)\n",
	     ttree->metric_dist, ttree->metric_cross,
	     ttree->metric_distx, id, str );
}

void
calc_print_tree_metrics( FILE *fp, tech_tree_t *ttree, int id, char *str )
{
    calc_tree_metrics( ttree );
    print_tree_metrics( fp, ttree, id, str );
}

/* must have already called calc_tree_metrics */
void
print_tree_metrics_ps( FILE *fp, tech_tree_t *ttree )
{
    fprintf( fp, "%% techtree metrics:\n" );
    fprintf( fp, "%%  pathlength = %g\n", ttree->metric_dist );
    fprintf( fp, "%%  crossings = %d\n", ttree->metric_cross );
    fprintf( fp, "%%  crossing weighted pathlength = %g\n",
	     ttree->metric_distx );
}
     


