/********************************************************************** 
 techtree - Copyright (C) 1998 - David W. Pfitzner
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "techtree.h"
#include "options.h"
#include "new_layout.h"
#include "my_vop.h"

static techtree_opt_t *opt;	/* for options from techtree.c */
static FILE *debug_out;		/* for debug_out from techtree.c */

/* debug levels, for opt->debug_new */
#define DEBUG_TERSE 5
#define DEBUG_ONE_DETAIL 6
#define DEBUG_EACH_ITER 8
#define DEBUG_EACH_DETAIL 9

/* could make these options some time: */

#define VSCALE 1		/* vertical distance = d(level)/NODE_SY  */
#define RMIN 1e-6
#define SAME_SIZE 0

#define FSMALL 1e-3
#define NITER_MAX 2000
#define NDIR_TRY 90

#define FREP_A 1

#define FATT_NX 1		/* 1,2 */
#define FATT_NY 1		/* 0,1,2 */
#define FATT_A 1

#define FCEN_N 1
#define FCEN_A 1 /* 0.01 */

#define FCOL_N 1
#define FCOL_A_INIT 1
#define FCOL_A_FACTOR 1.01
#define DMAX_STOP 0.01

#define ACC_EPS 0.2
#define GMAX_HIGH 0.75		/* less than 2 */
#define GMAX_LOW 0.25		/* 0.1 */
#define DX_INCREASE_FACTOR 2.0
#define DX_REDUCE_FACTOR 0.5

#define RAND_U01 (rand()/(float)RAND_MAX)

static tech_node_t *
follow_allow( tech_tree_t *ttree, tech_node_t *node)
{
    if( !node->pseudo ) {
	return node;
    }
    assert(node->n_allow==1);
    return follow_allow( ttree, TNODE_P(ttree,node->i_allow[0]) );
}

static tech_node_t *
follow_req( tech_tree_t *ttree, tech_node_t *node)
{
    if( !node->pseudo ) {
	return node;
    }
    assert(node->n_req==1);
    return follow_req( ttree, TNODE_P(ttree,node->i_req[0]) );
}

static void
one_linear_pseudo( tech_tree_t *ttree, tech_node_t *node)
{
    tech_node_t *n1, *n2;
    float f0, f1, f2;
    
    n1 = follow_req(ttree, node);
    n2 = follow_allow(ttree, node);
    f0 = 1.0/(float)(n2->level - n1->level);
    f1 = n2->level - node->level;
    f2 = node->level - n1->level;
    f1 *= f0;
    f2 *= f0;
    VVV(node->x, = f1*n1->x, + f2*n2->x );
}

static void
all_linear_pseudo( tech_tree_t *ttree )
{
    tech_node_t *node;
    int i;

    for( i=0, node=ttree->nodes; i<ttree->n; node++, i++ ) {
	if( node->pseudo ) {
	    one_linear_pseudo(ttree, node);
	}
    }
}

static void
print_newlayout( FILE *fp, tech_tree_t *ttree, int iter )
{
    tech_node_t *node, *a_node;
    int i, k;
    
    for( i=0, node=ttree->nodes; i<ttree->n; node++, i++ ) {
	fprintf( fp, "%g %g %d 0 0 -1 %d %s\n", node->x[0], node->x[1],
		node->level, iter, node->name );
    }
    for( i=0, node=ttree->nodes; i<ttree->n; node++, i++ ) {
	for( k=0; k<node->n_allow; k++ ) {
	    a_node = TNODE_P(ttree,node->i_allow[k]);
	    fprintf( fp, "%g %g %d %g %g %d %d link %s -> %s\n",
		    node->x[0], node->x[1], node->level,
		    a_node->x[0], a_node->x[1], a_node->level,
		    iter, node->name, a_node->name );
	}
    }
}

static void
initial_scatter( tech_tree_t *ttree )
{
    tech_node_t *node;
    float r_init, try_x[NDIM], r2;
    int i;

    r_init = sqrt(ttree->max_one_level);
    if( opt->debug_new >= DEBUG_TERSE ) {
	fprintf( debug_out, "# r_init = %g\n", r_init );
    }
    for( i=0, node=ttree->nodes; i<ttree->n; node++, i++ ) {
	VS( node->force, = 0.0 );
	VS( node->force_prev, = 0.0 );
    }
    for( i=0, node=ttree->nodes; i<ttree->n; node++, i++ ) {
	if( node->pseudo ) {
	    break;
	}
	do {
	    VS(try_x, = -1.0+2.0*RAND_U01);
	    r2 = Dot2(try_x);
	} while( r2 > 1.0 );
	VV( node->x, = r_init*try_x );
    }
    all_linear_pseudo(ttree);
}

static void
calc_forces( tech_tree_t *ttree, float *dir, float acol )
{
    tech_node_t *node, *n1, *n2;
    int i, n_this_level, j1, j2, k1, k2;
    float dx[NDIM], f[NDIM], r, invr, fmag, u, y, uy, usign;
    float c0, c1, c2;
    float major[NDIM];
    float level_r_min;
    float s1, s2;

    if( dir!=NULL ) {
	VeqVperp( major, dir );
    }
    
    /** zero initial forces: **/
    for( i=0, node=ttree->nodes; i<ttree->n; node++, i++ ) {
	VV( node->force_prev, = node->force );
	VS( node->force, = 0.0 );
    }
    /** repulsive forces at each level: **/
    level_r_min = 1e+20;
    for( i=0; i<ttree->n_levels; i++ ) {
	n_this_level = ttree->per_level[i];
	
	for( j1=0; j1<n_this_level-1; j1++ ) {
	    k1 = ttree->i_sort[ttree->level_start[i] + j1];
	    n1 = TNODE_P(ttree, k1);
	    s1 = TNODE_REL_WIDTH(ttree,n1);
	    
	    for( j2=j1+1; j2<n_this_level; j2++ ) {
		k2 = ttree->i_sort[ttree->level_start[i] + j2];
		n2 = TNODE_P(ttree, k2);
		s2 = TNODE_REL_WIDTH(ttree,n2);

		if( n1->pseudo && n2->pseudo && opt->frep_do_pp ) {
		  continue;
		}

		/* repulsive force between n1 and n2: */
		VVV(dx, = n1->x, - n2->x);
		r = sqrt(Dot2(dx));
		if( r < level_r_min ) {
		    level_r_min = r;
		}
		if( r<RMIN ) {
		    r = RMIN;
		}
		invr = 1.0/r;
		if( SAME_SIZE ) {
		    u = 2.0*invr;
		} else {
		    u = (s1+s2)*invr;
		}
		u *= opt->frep_size;
		if( u < 1.0 ) {
		    fmag = FREP_A * pow( u, opt->frep_n_inner );
		} else {
		    fmag = FREP_A * pow( u, opt->frep_n_outer );
		}
		VV(f, = (fmag*invr)*dx );
		VV(n1->force, += f);
		VV(n2->force, -= f);
	    }
	}
    }
    if( opt->debug_new >= DEBUG_EACH_ITER ) {
	fprintf( debug_out, "# level_r_min = %g\n", level_r_min );
    }
    
    /** forces along links: **/
    for( i=0, n1=ttree->nodes; i<ttree->n; n1++, i++ ) {
	if( opt->init_pseudo_linear && n1->pseudo ) {
	    break;
	}
	for( j1=0; j1<n1->n_allow; j1++ ) {
	    node = TNODE_P(ttree, n1->i_allow[j1]);
	    n2 = follow_allow(ttree,node);
	    
	    /* attractive force between n1 and n2: */
	    VVV(dx, = n1->x, - n2->x);
	    r = sqrt(Dot2(dx));
	    if( r<RMIN ) {
		r = RMIN;
	    }
	    invr = 1.0/r;
	    y = n2->level - n1->level;
	    uy = VSCALE/y;
	    fmag = FATT_A * pow( r, FATT_NX ) * pow( uy, FATT_NY );
	    VV(f, = (fmag*invr)*dx );
	    VV(n1->force, -= f);
	    VV(n2->force, += f);

	    /* forces on n1 and n2 from pseudo-nodes between: */
	    while( opt->init_pseudo_linear && node->pseudo ) {
		c0 = 1.0/(n2->level - n1->level);
		c1 = c0*(node->level - n1->level);
		c2 = c0*(n2->level - node->level);

		VV(n1->force, += c2*node->force);
		VV(n2->force, += c1*node->force);
		
		assert(node->n_allow==1);
		node = TNODE_P(ttree, node->i_allow[0]);
	    }
	}
    }

    /** global force: **/
    for( i=0, node=ttree->nodes; i<ttree->n; node++, i++ ) {
	r = sqrt(Dot2(node->x));
	if( r<RMIN ) {
	    r = RMIN;
	}
	invr = 1.0/r;
	fmag = FCEN_A * pow( r, FCEN_N );
	VV( node->force, -= (fmag*invr) * node->x );

	if( dir==0 ) {
	    continue;
	}

	u = Dot(node->x,dir);
	usign = (u>0?1:u<0?-1:0);

	fmag = acol * pow( (usign*u), FCOL_N );
	VV( node->force, -= fmag*usign*dir );
    }
}

/* below:
   g == |force-force_prev|/max(|force|,|force_prev|)
*/
static float
calc_force_max( tech_tree_t *ttree, float *gmax_return )
{
    tech_node_t *node;
    float dfv[NDIM];
    float f, fpre, df, g, mf;
    float gmax, fmax;
    int i, i_gmax=-1, i_fmax=-1;

    fmax = 0.0;
    gmax = 0.0;
    for( i=0, node=ttree->nodes; i<ttree->n; node++, i++ ) {
	if( opt->init_pseudo_linear && node->pseudo ) {
	    break;
	}
	f = sqrt(Dot2(node->force));
	fpre = sqrt(Dot2(node->force_prev));
	VVV(dfv, = node->force, - node->force_prev);
	df = sqrt(Dot2(dfv));
	mf = (f > fpre) ? f : fpre;
	if( i_fmax==-1 || f > fmax ) {
	    fmax = f;
	    i_fmax = i;
	}
	if( mf>0 ) {
	    /* (ignore nodes with both forces zero) */
	    g = df / mf;
	    if( i_gmax==-1 || g > gmax ) {
		gmax = g;
		i_gmax = i;
	    }
	}
    }
    if( opt->debug_new >= DEBUG_EACH_ITER ) {
	node = TNODE_P(ttree,i_fmax);
	fprintf( debug_out, "# force_max = %g (%s x %g %g f %g %g)\n",
		 fmax, node->name,
		 node->x[0], node->x[1],
		 node->force[0], node->force[1] );
	node = TNODE_P(ttree,i_gmax);
	fprintf( debug_out, "# g_force_max = %g (%s f %g %g fpre %g %g)\n",
		 gmax, node->name,
		 node->force[0], node->force[1],
		 node->force_prev[0], node->force_prev[1] );
    }
    if( gmax_return != NULL ) {
	*gmax_return = gmax;
    }
    return fmax;
}

static float
calc_dperp_max( tech_tree_t *ttree, float *dir )
{
    tech_node_t *node;
    float d, dmax;
    int i;

    assert(dir!=NULL);

    dmax = 0;
    for( i=0, node=ttree->nodes; i<ttree->n; node++, i++ ) {
	d = fabs(Dot(node->x,dir));
	if( d > dmax ) {
	    dmax = d;
	}
    }
    if( opt->debug_new >= DEBUG_EACH_ITER ) {
	fprintf( debug_out, "# dperp_max = %g\n", dmax );
    }
    return dmax;
}

/* dmax<0 signals not using collapse to dir */
static void
apply_forces( int iter, tech_tree_t *ttree, float fmax, float dmax, float gmax )
{
    static float target_dx_last = 0.0;

    tech_node_t *node;
    float target_dx_max = ACC_EPS;
    float target_dx, alpha;
    int i;

    assert(fmax>0);
    if( dmax >= 0 ) {
	target_dx = ACC_EPS * ((dmax < 1.0) ? dmax : 1.0);
    } else if( iter==0 ) {
	target_dx = ACC_EPS;
    } else if( gmax > GMAX_HIGH ) {
	target_dx = target_dx_last * DX_REDUCE_FACTOR;
    } else if( gmax < GMAX_LOW ) {
	target_dx = target_dx_last * DX_INCREASE_FACTOR;
    } else {
	target_dx = target_dx_last;
    }

    if( target_dx > target_dx_max ) {
	target_dx = target_dx_max;
    }
    alpha = target_dx / fmax;
    
    if( opt->debug_new >= DEBUG_EACH_ITER ) {
	fprintf( debug_out, "# target_dx = %g, alpha = %g\n",
		 target_dx, alpha );
    }
    target_dx_last = target_dx;
    
    for( i=0, node=ttree->nodes; i<ttree->n; node++, i++ ) {
	if( opt->init_pseudo_linear && node->pseudo ) {
	    break;
	}
	VV(node->x, += alpha*node->force);
    }
    if( opt->init_pseudo_linear ) {
	all_linear_pseudo(ttree);
    }
}

/* If dir==0, use fixed central force; else growing anisotropic force
   to collapse along that direction
*/
static void
evolve_to_static( tech_tree_t *ttree, float *dir )
{
    int i = 0;
    float acol = (dir?FCOL_A_INIT:0.0);
    float fmax = 0.0;
    float gmax = 0.0;
    float dmax = -1.0;

    for(;;) {
	if( opt->debug_new >= DEBUG_EACH_ITER ) {
	    fprintf( debug_out, "# i %d acol %g\n", i, acol );
	}
	calc_forces( ttree, dir, acol );
	fmax = calc_force_max( ttree, &gmax );
	if( dir==NULL ) {
	    if( fmax < FSMALL ) {
		if( opt->debug_new >= DEBUG_TERSE ) {
		    fprintf( debug_out, "# breaking on fmax = %g\n", fmax );
		}
		break;
	    }
	} else {
	    dmax = calc_dperp_max( ttree, dir );
	    if( dmax < DMAX_STOP ) {
		if( opt->debug_new >= DEBUG_TERSE ) {
		    fprintf( debug_out, "# breaking at dmax_stop (dmax = %g)\n",
			     dmax );
		}
		break;
	    }
	}
	if( i>=NITER_MAX ) {
	    if( opt->debug_new >= DEBUG_TERSE ) {
		fprintf( debug_out, "# breaking at itermax (last fmax %g)\n",
			 fmax );
	    }
	    break;
	}
	apply_forces( i, ttree, fmax, dmax, gmax );
	i++;
	if( dir!=NULL ) {
	    acol *= FCOL_A_FACTOR;
	}
	if( opt->debug_new >= DEBUG_EACH_DETAIL ) {
	    print_newlayout( debug_out, ttree, i*(dir==NULL?1:-1) );
	}
    }
    if( opt->debug_new >= DEBUG_ONE_DETAIL ) {
	print_newlayout( debug_out, ttree, (i+1)*(dir==NULL?1:-1) );
    }
}

/* sum of projected horizontal link distances */
static float
calc_dir_proj( tech_tree_t *ttree, float *dir )
{
    tech_node_t *node, *a_node;
    float major[NDIM], link[NDIM], proj, sum=0.0;
    int i, k;

    VeqVperp( major, dir );
    
    for( i=0, node=ttree->nodes; i<ttree->n; node++, i++ ) {
	for( k=0; k<node->n_allow; k++ ) {
	    a_node = TNODE_P(ttree,node->i_allow[k]);
	    VVV( link, = node->x, - a_node->x );
	    proj = fabs(Dot(link,major));
	    sum += proj;
	}
    }
    return sum;
}

/* dir is direction of "minor axis" to collapse to zero */
static void
choose_collapse_dir( tech_tree_t *ttree, float *dir )
{
    float pi, ang;
    float best_dir[NDIM];
    float proj, best_proj=-1, best_ang=-1;
    int i;

    for( i=0; i<NDIR_TRY; i++ ) {
	pi = 4.0*atan(1.0);
	ang = pi*(i/(float)NDIR_TRY);
	dir[0] = cos(ang);
	dir[1] = sin(ang);
	proj = calc_dir_proj( ttree, dir );
	if( opt->debug_new >= DEBUG_ONE_DETAIL ) {
	    fprintf( debug_out, "# i %d ang %g proj %g\n", i, ang, proj );
	}
	if( best_proj==-1 || proj < best_proj ) {
	    best_ang = ang;
	    best_proj = proj;
	    VV(best_dir, = dir);
	}
    }
    VV(dir, = best_dir);
    if( opt->debug_new >= DEBUG_TERSE ) {
	fprintf( debug_out, "# collapse dir %g %g (ang %g)\n",
		 dir[0], dir[1], best_ang );
    }
}

static void 
collapse_final( tech_tree_t *ttree, float *dir )
{
    tech_node_t *node;
    float major[NDIM], dx_min=0, dx_max=0, f;
    float vert, vert_max=0;
    int i;

    if( opt->init_pseudo_linear ) {
	all_linear_pseudo(ttree);
    }
    VeqVperp( major, dir );
    for( i=0, node=ttree->nodes; i<ttree->n; node++, i++ ) {
	node->dx = Dot(node->x,major);
	if( i==0 || (node->dx < dx_min) ) {
	    dx_min = node->dx;
	}
	if( i==0 || (node->dx > dx_max) ) {
	    dx_max = node->dx;
	}
	vert = Dot(node->x,dir);
	if( vert > vert_max ) {
	    vert_max = vert;
	}
    }
    
    if( opt->debug_new >= DEBUG_TERSE ) {
	fprintf( debug_out, "# collapse dx_min = %g, dx_max = %g,"
		 " vert_max = %g\n", dx_min, dx_max, vert_max );
    }
    
    f = dx_max - dx_min;
    assert(f!=0);
    f = 1.0/f;
    f /= (1.0+ttree->one_wide);
    
    for( i=0, node=ttree->nodes; i<ttree->n; node++, i++ ) {
	node->dx = 0.5*ttree->one_wide + (node->dx - dx_min)*f;
    }
}

/* entry function: calc dx values */
void
calc_new_layout( tech_tree_t *ttree, techtree_opt_t *options,
		 FILE *debug_out_fp )
{
    float dir[NDIM];

    opt = options;
    debug_out = debug_out_fp;
    
    initial_scatter( ttree );
    evolve_to_static( ttree, NULL );
    choose_collapse_dir( ttree, dir );
    evolve_to_static( ttree, dir );
    collapse_final( ttree, dir );
}


