/********************************************************************** 
 techtree - Copyright (C) 1998 - David W. Pfitzner
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "techtree.h"
#include "options.h"
#include "metrics.h"
#include "postscript.h"
#include "city.h"
#include "unit.h"
#include "extra.h"

static techtree_opt_t *opt;	/* for pointer to options from techtree.c */

/* options not fluid enough to have been put in opt: */
#define FRAC_CORNER 0.5		/* for curved path corners */
#define BOX_SZ 3		/* boxes at nodes, in ps units */
#define BOX_SZ_SMALL 2
#define FONT_SZ 8.5		/* tech node labels */
#define FONT_SZ_SMALL 5.5	/* to write small */
#define EXTRA_BOTTOM 25		/* for credits line, in postscript units */

/* debug levels, for opt->debug_ps */
#define DEBUG_TERSE 1
#define DEBUG_EACHTECH 2

/* a bounding box: */
typedef struct {
    int x1, x2, y1, y2;
} bbox_t;

/* Calculate bounding box on physical page, in postscript units,
   given margin specs etc.  Because we don't really know about font
   sizes etc, we can't guaranttee that labels etc will stay inside
   this.  (Except we will use clipping).  
*/
static void
calc_bbox_page( bbox_t *bb ) {
    float bbx1, bbx2, bby1, bby2; /* bounding box in mm*/
    float one_mm_in_ps;

    one_mm_in_ps = 0.1*72.0/2.54;

    bbx1 = opt->page_margin;
    bbx2 = opt->page_width - opt->page_margin;
    bby1 = opt->page_margin;
    bby2 = opt->page_height - opt->page_margin;

    bb->x1 = floor(bbx1*one_mm_in_ps);
    bb->x2 = ceil(bbx2*one_mm_in_ps);
    bb->y1 = floor(bby1*one_mm_in_ps);
    bb->y2 = ceil(bby2*one_mm_in_ps);
}

/* Actually just print postscript to _define_ a title/credits line;
   Note here bb is real page bbox */
static void
print_ps_title_cred( bbox_t *bb )
{
    int x, y;

    x = bb->x1 + 0.5*(bb->x2 - bb->x1);
    y = bb->y1 + 5;

    /* Could be a different font/size: */
    printf( "/my_credits {\n" );
    printf( "  gsave\n" );
    printf( "  0 setgray\n" );
    printf( "  /Helvetica findfont\n" );
    printf( "  %g scalefont setfont\n", FONT_SZ );
    printf( "  %d %d translate\n", x, y );
    printf( "  (FreeCiv Technology Tree, " );
    printf( "by techtree v%s, ", version );
    printf( "by David W. Pfitzner, " );
    printf( "dwp@mso.anu.edu.au) S2a\n" );
    printf( "  grestore\n" );
    printf( "} def\n\n" );
}

/* Print postscript header section */
static void
print_ps_header( bbox_t *bb, tech_tree_t *ttree )
{
    float node_font_sz;
    int box_sz;

    node_font_sz = opt->tech_info ? FONT_SZ_SMALL : FONT_SZ;
    box_sz = opt->tech_info ? BOX_SZ_SMALL : BOX_SZ;

    printf( "%%!PS-Adobe-2.0\n" );
    /* or: printf( "%%!PS-Adobe-2.0 EPSF-2.0"; */
    printf( "%%%%Creator: techtree version %s by David W. Pfitzner\n", version );
    printf( "%%%%BoundingBox: %d %d %d %d\n", bb->x1, bb->y1, bb->x2, bb->y2 );
    /* printf( "%%%%Pages: 1\n" ); */
    printf( "%%%%Pages: %d\n", opt->two_page ? 2 : 1 ); 
    printf( "%%%%DocumentFonts: Helvetica\n" );
    printf( "%%%%EndComments\n\n" );
    
    print_options( stdout, opt, version, "% " );
    printf( "\n" );
    print_tree_metrics_ps( stdout, ttree );
    printf( "\n" );

    printf( "save                 %% remember original state\n" );
    printf( "40 dict begin        %% a local dictionary\n" );
    printf( "\n" );

    printf( "/D { 1 index 1 index } def\n"
	    "  %% (x y) duplicate to (x y x y)\n" );
    printf( "/L { newpath moveto lineto stroke } def\n"
	    "  %% (x1 y1 x2 y2) line\n" );
    printf( "/L2 { /y exch def /x exch def x y L x y } def\n"
	    "  %% (x1 y1 x2 y2) line, with x2, y2 back on stack at end\n" );
    printf( "/C1 { newpath moveto } def\n"
	    "  %% (x1 y1) start curve\n" );
    printf( "/C2 { D } def\n"
	    "  %% (x2 y2) apex of curve\n" );
    printf( "/C3 { /y exch def /x exch def x y curveto stroke x y } def\n"
	    "  %% (x3 y3) end of curve, leaving x3, y3 on stack\n" );
    printf( "/B { gsave translate \n"
	    "  %d dup dup scale 1 exch div setlinewidth\n"
	    "  newpath -1 -1 moveto -1 1 lineto 1 1 lineto 1 -1 lineto\n"
	    "  closepath stroke grestore } def\n" /* or stroke->fill */
	    "  %% (x y) box\n", box_sz );
    printf( "/P { pop pop } def  \n"
	    "  %% (x y) pop coords\n" );
    printf( "/S1 { D gsave translate 0 -%g translate } def\n"
	    "  %% (x y) setup for label, leave (x y) on stack\n",
	    (node_font_sz+box_sz) );
    printf( "/S2a { gsave newpath 0 0 moveto dup\n"
	    "  stringwidth pop 2 div neg 0 rmoveto show grestore } def\n"
	    "  %% (string) write centered string\n" );
    printf( "/S2 { S2a 0 -%g translate } def\n"
	    "  %% (string) write centered string and go down a bit\n",
	    (node_font_sz+0*(opt->tech_info>0)) );
    printf( "/S3 { grestore } def\n"
	    "  %% after finished label\n" );
    printf( "/G1 { gsave [ 6 3 ] 0 setdash } def\n"
	    "  %% start dashed line (was gray line)\n" );
    printf( "/G2 { grestore } def\n"
	    "  %% end dashed line (was gray line)\n" );
    printf( "\n" );

    printf( "/BB {\n" );
    printf( "  /y2 exch def /y1 exch def /x2 exch def /x1 exch def\n" );
    printf( "  newpath x1 y1 moveto\n" );
    printf( "  x2 y1 lineto\n" );
    printf( "  x2 y2 lineto\n" );
    printf( "  x1 y2 lineto\n" );
    printf( "  closepath\n" );
    printf( "} def   %% (x1) (x2) (y1) (y2) define a bbox path\n" );

    printf( "/my_cliptopage {\n" );
    printf( "  %d %d %d %d BB\n", bb->x1, bb->x2, bb->y1, bb->y2 );
    printf( "  %% gsave stroke grestore %% uncomment to show bounding box\n" );
    printf( "  clip\n" );
    printf( "} def\n\n" );

    printf( "/my_defaultfont {\n" );
    printf( "  /Helvetica findfont   %% Get the basic font\n" );
    printf( "  %g scalefont           %% Scale the font to X points\n",
	    node_font_sz );
    printf( "  setfont               %% Make it the current font\n" );
    printf( "} def\n\n" );

    printf( "/links_follow {\n" );
    printf( "  %g setlinewidth\n", 0.01*opt->line_width );
    printf( "  %g setgray  \n", 0.01*opt->line_gray );
    printf( "} def   %% thin grey lines (text over)\n\n" );

    printf( "/nodes_follow {\n" );
    printf( "  my_defaultfont\n" );
    printf( "  0 setgray  \n" );
    printf( "} def   %% black boxes/text\n\n" );

    print_ps_title_cred( bb );
	
    printf( "%%%%EndProlog\n\n" );
}

static void
print_ps_endpart( void )
{
    printf( "%%%%Trailer\n\n" );

    printf( "end                  %% stop using temporary dictionary\n" );
    printf( "restore              %% restore original state\n\n" );
    printf( "%%%%EOF\n" );
}

static void
dxy_to_ps( bbox_t *bb, float *dx, float *dy )
{
    *dx = bb->x1 + (*dx)*(bb->x2-bb->x1);
    *dy = bb->y1 + (*dy)*(bb->y2-bb->y1);
}

static void
node_dx_dy( bbox_t *bb, tech_tree_t *ttree, int i_node, float *dx, float *dy )
{
    tech_node_t *node = TNODE_P(ttree,i_node);
    *dx = node->dx;
    *dy = node->dy;
    if( !(opt->first_at_top) ) {
	*dy = 1 - *dy;
    }
    dxy_to_ps( bb, dx, dy );
}

/* strcpy and convert (sometimes) to lowercase */
static void
lower_str( char *snew, char *sold, int nmax )
{
    int i;
    
    strncpy( snew, sold, nmax );
    for( i=0; i<nmax; i++ ) {
	if( snew[i] == '\0' ) {
	    break;
	}
	/* Only lower if the next letter is lower-case;
	   otherwise its probably an acronym, or barracks II etc */
	if( snew[i+1] != '\0' && islower(snew[i+1]) ) {
	    snew[i] = tolower(snew[i]);
	}
    }
}

static void
print_tech_info( tech_node_t *node )
{
    char buf[256];
    int target_tech = node->orig_itech;
    int i, j, n_obs_u;

    buf[sizeof(buf)-1] = '\0';

    if( extras[target_tech] != NULL ) {
	printf( "((%s)) S2\n", extras[target_tech] );
    }
    
    /* this could be more efficient... */
    for( i=0; i<U_LAST; i++ ) {
	if( unit_types[i].tech_requirement == target_tech ) {
	    lower_str( buf, unit_types[i].name, sizeof(buf)-1 );
	    printf( "(%s) S2\n", buf );
	}
    }
    for( i=0; i<B_LAST; i++ ) {
	if( improvement_types[i].tech_requirement == target_tech ) {
	    if( improvement_types[i].is_wonder ) {
		strncpy( buf, improvement_types[i].name, sizeof(buf)-1 );
	    } else {
		lower_str( buf, improvement_types[i].name, sizeof(buf)-1 );
	    }
	    printf( "(%s) S2\n", buf );
	}
    }
    n_obs_u = 0;
    for( i=0; i<U_LAST; i++ ) {
	if( (j=unit_types[i].obsoleted_by) != -1 &&
	    unit_types[j].tech_requirement == target_tech ) {
	    lower_str( buf, unit_types[i].name, sizeof(buf)-1 );
	    if( n_obs_u == 0 ) {
		printf( "(~%s", buf );
	    } else {
		printf( ", %s", buf );
	    }
	    n_obs_u++;
	}
    }
    if( n_obs_u > 0 ) {
	printf( ") S2\n" );
    }
    for( i=0; i<B_LAST; i++ ) {
	if( improvement_types[i].obsolete_by == target_tech ) {
	    if( improvement_types[i].is_wonder ) {
		strncpy( buf, improvement_types[i].name, sizeof(buf)-1 );
	    } else {
		lower_str( buf, improvement_types[i].name, sizeof(buf)-1 );
	    }
	    printf( "(~%s) S2\n", buf );
	}
    }
}

static void
start_page( bbox_t *bb, int n )
{
    float vscale, hscale, xs, ys, f;

    assert(n==1 || n==2);
    printf( "%%%%Page: %d %d\n\n", n, opt->two_page ? 2 : 1 );
    
    printf( "gsave\n" );
    printf( "my_cliptopage\n\n" );
    
    printf( "%% %d %d %d %d BB stroke\n\n", bb->x1, bb->x2, bb->y1, bb->y2 );

    if( opt->two_page == 0 ) {
	return;
    }

    f = (100.0 + opt->two_overlap)/200.0;
    vscale = (bb->y2 - bb->y1) / (float)(bb->x2 - bb->x1);
    hscale = (bb->x2 - bb->x1) / (f*(float)(bb->y2 - bb->y1));
    ys = bb->y1 - bb->x1*vscale;
    if( n==1 ) {
	xs = bb->x1 + bb->y2*hscale;
    } else {
	xs = bb->x2 + bb->y1*hscale;
    }
    printf( "gsave   %% setting up for sideways over two pages, page %d\n", n );
    printf( "%g %g translate\n", xs, ys );
    printf( "90 rotate\n" );
    printf( "%g %g scale\n", vscale, hscale );
    printf( "\n" );
    printf( "%% %d %d %d %d BB stroke\n\n", bb->x1, bb->x2, bb->y1, bb->y2 );
}

static void
end_page( void ) {
    if( opt->two_page ) {
	printf( "grestore   %% translate/rotate\n" );
    }
    printf( "my_credits\n\n" );
    printf( "grestore   %% my_cliptopage\n" );
    printf( "showpage\n\n" );
}

static int
follow_allow_level( tech_tree_t *ttree, tech_node_t *node, int k_allow )
{
    tech_node_t *a_node = TNODE_P(ttree,node->i_allow[k_allow]);
    
    if( a_node->pseudo ) {
	assert(a_node->n_allow==1);
	return follow_allow_level(ttree,a_node,0);
    } else {
	return a_node->level;
    }
}
     
static void
print_ps_nodes( tech_tree_t *ttree, bbox_t *bb, int page )
{
    tech_node_t *node, *lnode, *nnode;
    int i, j, k;
    float dx, dy, dx1, dy1, dx2, dy2;
    char buf[128], *tok;
    int this_islower, partial;
    int crit_level = 0;
    int allows_go_off_page = 0;
    int on_stack;
    float fcrit;

    assert(page==1 || page==2);

    if( opt->two_page ) {
	if( (opt->first_at_top && page==1) ||
	    (!opt->first_at_top && page==2) ) {
	    allows_go_off_page = 1;
	} else {
	    allows_go_off_page = 0;
	}
	fcrit = 0.5*ttree->n_levels;
	if( allows_go_off_page ) {
	    /* crit to higher levels: */
	    fcrit += (opt->two_overlap/100.0)*(0.5*ttree->n_levels);
	    crit_level = floor(fcrit);
	} else {
	    /* crit to lower levels: */
	    fcrit -= (opt->two_overlap/100.0)*(0.5*ttree->n_levels);
	    crit_level = floor(fcrit);
	}
	if( opt->debug_ps >= DEBUG_TERSE ) {
	    printf( "%% allows_go_off_page=%d crit_level=%d "
		    "(fcrit=%g n_levels=%d)\n",
		    allows_go_off_page, crit_level,
		    fcrit, ttree->n_levels );
	}
    }

    /* links; do these first so text goes over top: */
    
    printf( "links_follow\n" );
    
    for( i=0; i<ttree->n; i++ ) {
	j = ttree->i_sort[i];
	node = ttree->nodes + j;
	if( opt->two_page && allows_go_off_page && node->level>crit_level ) {
	    continue;
	}
	node_dx_dy( bb, ttree, j, &dx, &dy );

	if( !node->pseudo ) {
	    on_stack = 0;
	    for( k=0; k<node->n_allow; k++ ) {
		if( opt->two_page && (!allows_go_off_page)
		    && node->level < crit_level
		    && follow_allow_level(ttree,node,k) < crit_level ) {
		    continue;
		}
		if( !on_stack ) {
		    /* put this coord on stack: */
		    printf( "%d %d\n", (int)dx, (int)dy );
		    on_stack = 1;
		    if( opt->debug_ps >= DEBUG_EACHTECH ) {
			printf( "%% (%s)\n", node->name );
		    }
		}
		/* duplicate node coord so it stays on stack for later */
		printf( "D " );
		/* In loop below, we're drawing a line as
		   part of a allow link.  On that stack are (x,y) of
		   where the line is up to so far.  We are to use
		   those and remove them from the stack.
		   Once lnode is set, it point to the next node in the
		   path, which could be pseudo (keep going) or real (stop).
		   dx2 and dy2 are are the x and y of previous node in path
		   which may be same or different from stack values...
		   For lnode real, just do a straight line to the
		   endpoint.  For lnode pseudo, draw a line to the
		   near the corner, then a curve.
		*/
		if( node->redundant_allow[k] ) {
		    printf( "G1 " );
		}
		nnode =	TNODE_P(ttree, node->i_allow[k]);
		dx2 = dx;
		dy2 = dy;
		do {
		    lnode = nnode;
		    node_dx_dy( bb, ttree, TNODE_I(ttree,lnode),
				&dx1, &dy1 );
		    if( opt->two_page && allows_go_off_page
			&& lnode->level>crit_level+1 ) {
			printf( "P\n" );
			break;
		    } 
		    if( !lnode->pseudo ) {
			/* finish the path: */
			printf( "%d %d L\n", (int)dx1, (int)dy1 );
		    } else {
			/* to start of corner: */
			dx2 = dx2 + (1-FRAC_CORNER)*(dx1-dx2);
			dy2 = dy2 + (1-FRAC_CORNER)*(dy1-dy2);
			printf( "%d %d L2\n", (int)dx2, (int)dy2 );
			/* through corner: need next node */
			nnode = TNODE_P(ttree, lnode->i_allow[0]);
			node_dx_dy( bb, ttree, TNODE_I(ttree,nnode),
				    &dx2, &dy2 );
			dx2 = dx1 + FRAC_CORNER*(dx2-dx1);
			dy2 = dy1 + FRAC_CORNER*(dy2-dy1);
			printf( "C1 %d %d C2 %d %d C3\n",
				(int)dx1, (int)dy1, (int)dx2, (int)dy2 );
		    }
		    dx2 = dx1;
		    dy2 = dy1;
		} while( lnode->pseudo );
		if( node->redundant_allow[k] ) {
		    printf( "G2\n" );
		}
	    }
	    if( on_stack ) {
		printf( "P\n" );
	    }
	}
    }
    
    printf( "nodes_follow\n" );

    if( 0 ) {			/* probably paranoia */
	/* node labels may "spill over" to space associated with lower nodes: */
	if( opt->two_page && page==2 ) {
	    if( opt->first_at_top ) {
		crit_level--;
	    } else {
		crit_level++;
	    }
	    if( opt->debug >= DEBUG_TERSE ) {
		printf( "%% crit_level now %d\n", crit_level );
	    }
	}
    }
    
    for( i=0; i<ttree->n; i++ ) {
	j = ttree->i_sort[i];
	node = ttree->nodes + j;
	if( opt->two_page &&
	    ( (allows_go_off_page && node->level>crit_level)
	      || ((!allows_go_off_page) && node->level<crit_level) ) ) {
	    continue;
	}
	
	/* box and name: */
	if( (!node->pseudo) || opt->pseudo_labels ) {
	    node_dx_dy( bb, ttree, j, &dx, &dy );
	    printf( "%d %d S1 \n", (int)dx, (int)dy );
	    if( opt->tech_info ) {
		/* Don't multi-line tech names */
		printf( "(%s:) S2\n", node->name );
		print_tech_info( node );
	    } else {
		strncpy( buf, node->name, sizeof(buf)-1 );
		buf[sizeof(buf)-1] = 0;
		tok = strtok( buf, " " );
		partial = 0;
		do {
		    this_islower = (tolower(tok[0]) == tok[0]);
		    printf( "%s%s%s",
			    (partial?" ":"("), tok,
			    (this_islower?"":") S2 ") );
		    partial = this_islower;
		} while( (tok = strtok( NULL, " " )) );
		if( partial ) {
		    printf( ") S2 " );
		}
	    }
	    if( node->pseudo ) {
		printf( "S3 P\n" );
	    } else {
		printf( "S3 B\n" );
	    }
	}
    }
    printf( "\n" );
}

/*** entry function: ***/
void
print_ps( tech_tree_t *ttree, techtree_opt_t *options )
{
    bbox_t bbox_page, bbox_nodes, bbox_nshift;
    int num_pages, i_page, shift_dy;

    opt = options;
    calc_bbox_page( &bbox_page );
    print_ps_header( &bbox_page, ttree );
    
    /* allow a bit at the bottom for the title/credits */
    bbox_nodes = bbox_page;
    bbox_nodes.y1 += EXTRA_BOTTOM;
    
    /* move everything up a bit, because nodes are centered,
       but labels drop below nodes */
    shift_dy = (bbox_nodes.y2 - bbox_nodes.y1)/(3*ttree->n_levels);
    bbox_nshift = bbox_nodes;
    bbox_nshift.y1 += shift_dy;
    bbox_nshift.y2 += shift_dy;

    num_pages = (opt->two_page ? 2 : 1);
    for( i_page=1; i_page<=num_pages; i_page++ ) {
	start_page( &bbox_nodes, i_page );
	print_ps_nodes( ttree, &bbox_nshift, i_page );
	end_page();
    }
    print_ps_endpart();
}
