/********************************************************************** 
 techtree - Copyright (C) 1998 - David W. Pfitzner
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#ifdef HAVE_CONFIG_H
#include <config.h>		/* VERSION */
#endif

#include "techtree.h"
#include "convert.h"
#include "ilist.h"
#include "util.h"
#include "sort.h"
#include "print.h"
#include "postscript.h"
#include "redundant.h"
#include "checkalloc.h"
#include "metrics.h"
#include "options.h"
#include "new_layout.h"
#include "extra.h"
#include "ruleset.h"

char *version = VERSION;
static techtree_opt_t options, *opt = &options;

/* Some less fluid compile-time options: */

/* Swapping adjacent techs to iteratively improve topology: */
#define SWAP_ON_DIST 2		/* metric to use for evaluating swaps:
				   1=distances, 0=crossings, 2=cross*dist */
#define SWAP_RAND_REV_ORDER 1	/* randomly reverse orders to try swaps */

float dist_dy_factor = 0.5;	/* for distances; used in metrics.c */

/* "Badswap" = doing a locally bad swap to try for better global topology:
   Chance of doing a badswap is p = A exp(-K*d^N),
   where d = metric difference between swapped and unswapped configs;
   A,N are parameters, and K is adjusted to try and swap some
   given fraction (F) of badswap candidates, which changes linearly
   as we iterate.
*/
#define BADSWAP_A 0.5		/* 0 for no badswaps, else 0.5 sensible? */
#define BADSWAP_N 1		/* N in above */
#define BADSWAP_F_INITIAL 0.1	/* fraction of "bad" swaps to do initially */
#define BADSWAP_F_FINAL 0.0	/* fraction at end (linear progression) */
#define BADSWAP_K_INITIAL 1	/* initial K in above */
#define BADSWAP_K_FACTOR 1.1	/* should be > 1; adjust K by this fraction */
#define BADSWAP_K_MAX 0		/* 0 means no max; >0 necessary for DIST=0 ? */

/* Relatively obscure: */
#define ITERSMOOTH_UNDERSHOOT 0.7 /* should be <= 1 */
#define ITERSMOOTH_REORDER 0	/* 1 = allow itersmooth to re-order */
#define ITER_SELF_WEIGHT 0.01


/* A random float uniform between 0 and 1: */
#define RAND_U01 (rand()/(float)RAND_MAX)

/* Current target F and currently used K for badswaps: */
static float badswap_f = BADSWAP_F_INITIAL;
static float badswap_k = BADSWAP_K_INITIAL;


FILE *debug_out;

/* Debug meanings: debug output for each with <= debug.
   If no debugging, postscript goes to stdout.
   If no postscript, debugging goes to stdout.
   If both, postscript to stdout, debug to stderr.
*/
#define DEBUG_FINAL_METRICS 1	/* metrics at end */
#define DEBUG_OPT_FINAL 2	/* options, and metric and end */
#define DEBUG_PROG_METRICS 3	/* iterative metrics */
#define DEBUG_SWAP_STATS 4	/* good and bad swap numbers */
#define DEBUG_SWAP_STATS2 5	/* adjustments of badswap_k */
#define DEBUG_DETAIL 10		/* 1-off details */
#define DEBUG_DXFLEX 12	
#define DEBUG_SWAP_EACH 19	/* random chance for each swap */


/************************* FUNCTIONS: ***********************/

/* calculate n_allow and i_allow[] for each node */
static void
calc_allow( tech_tree_t *ttree )
{
    tech_node_t *node, *req_node;
    int i, j, req;
    int max_allow = 0;
    
    for( i=0, node=ttree->nodes; i<ttree->n; i++, node++ ) {
	for( j=0; j<node->n_req; j++ ) {
	    req = node->i_req[j];
	    req_node = &ttree->nodes[req];
	    if( req_node->n_allow == MAX_LINKS ) {
		fprintf( stderr, "n_allow to large: \"%s\"\n",
			 req_node->name );
		print_full_tree( stderr, ttree );
		exit(1);
	    }
	    req_node->i_allow[req_node->n_allow] = i;
	    req_node->n_allow++;
	    if( req_node->n_allow > max_allow ) {
		max_allow = req_node->n_allow;
	    }
	}
    }
    if( opt->debug >= DEBUG_DETAIL ) {
	fprintf( debug_out, "# max_allow = %d\n", max_allow );
    }
}

/* calculate "level" for each node, using level_req[], n_level_req;
   also calculate ttree->n_levels */
static void
calc_levels( tech_tree_t *ttree )
{
    tech_node_t *node, *a_node;
    ilist_t stack;		/* node indices */
    int i, j, k, level, max_level = 0;

    /* init level zero nodes, and put on stack: */
    ilist_new( &stack, 10 );
    for( i=0, node=ttree->nodes; i<ttree->n; i++, node++ ) {
	node->n_level_req = 0;
	if( node->n_req == 0 ) {
	    node->level = 0;
	    ilist_push( &stack, i );
	}
    }
    /* for each tech on stack, fill in levels for its "allow"
       techs; when all req levels filled in, push new tech to stack */
    while( !ilist_empty(&stack) ) {
	i = ilist_pop(&stack);
	node = ttree->nodes + i;
	for( j=0; j<node->n_allow; j++ ) {
	    a_node = ttree->nodes + node->i_allow[j];
	    a_node->level_req[a_node->n_level_req] = node->level;
	    a_node->n_level_req++;
	    if( a_node->n_level_req == a_node->n_req ) {
		level = -1;
		for( k=0; k<a_node->n_req; k++ ) {
		    if( a_node->level_req[k] > level ) {
			level = a_node->level_req[k];
		    }
		}
		a_node->level = level + 1;
		if( a_node->level > max_level ) {
		    max_level = a_node->level;
		}
		ilist_push( &stack, node->i_allow[j] );
	    }
	}
    }
    ilist_free( &stack );
    
    /* check everything done ok: */
    for( i=0, node=ttree->nodes; i<ttree->n; i++, node++ ) {
	if( node->n_level_req != node->n_req ) {
	    print_full_tree( stderr, ttree );
	    fprintf( stderr, "bad calc_levels\n" );
	    exit(1);
	}
    }
    if( opt->debug >= DEBUG_DETAIL ) {
	fprintf( debug_out, "# max_level = %d\n", max_level );
    }
    ttree->n_levels = max_level + 1;
}

/* pseudo techs have no (real) name, and exactly one req and
   one allow; they are introduced so that all allow/req
   links are between neighbouring levels.
*/

/* add pseudo techs corresponding to 1 real link
   "to" is an allow of "from"
*/
static void
add_one_pseudo_tech( tech_tree_t *ttree, int from, int to )
{
    tech_node_t *node;
    int from_level, to_level;
    int n_old, n_extra, n_new;
    int i, i1, j;

    from_level = ttree->nodes[from].level;
    to_level = ttree->nodes[to].level;
    n_extra = to_level - from_level -1;
    n_old = ttree->n;
    n_new = n_old + n_extra;
    
    for( i=0; i<n_extra; i++ ) {
	i1 = n_old + i;
	node = techtree_addnode( ttree );
	assert( i1 == TNODE_I(ttree,node) );
	techtree_initnode( node );
	node->pseudo = 1;
	node->name = malloc(10);
	CHECK_ALLOC( node->name );
	sprintf( node->name, "%d", i1 );
	node->n_req = node->n_allow = 1;
	if( i==0 ) {
	    node->i_req[0] = from;
	} else {
	    node->i_req[0] = i1-1;
	}
	if( i==n_extra-1 ) {
	    node->i_allow[0] = to;
	} else {
	    node->i_allow[0] = i1+1;
	}
	node->level = from_level + 1 + i;
    }
    /* alter appropriate req/allow in from and to: */
    node = ttree->nodes + from;
    for( j=0; j<node->n_allow; j++ ) {
	if( node->i_allow[j] == to ) {
	    node->i_allow[j] = n_old;
	    break;
	}
    }
    node = ttree->nodes + to;
    for( j=0; j<node->n_req; j++ ) {
	if( node->i_req[j] == from ) {
	    node->i_req[j] = n_new-1;
	    break;
	}
    }
}

static void
add_pseudo_techs( tech_tree_t *ttree )
{
    tech_node_t *node, *a_node;
    int i, j, a_i, level_diff, n_orig;

    /* Careful here: the tech tree is changing underneath
       us as we add more nodes (eg: may be re-alloced)
    */
    n_orig = ttree->n;
    for( i=0; i<n_orig; i++ ) {
	node = TNODE_P(ttree,i);
	for( j=0; j<node->n_allow; j++ ) {
	    a_i = node->i_allow[j];
	    a_node = ttree->nodes + a_i;
	    level_diff = a_node->level - node->level;
	    assert( level_diff >= 1 );
	    if( level_diff > 1 ) {
		add_one_pseudo_tech( ttree, i, a_i );
		/* recalc the pointer after possible realloc: */
		node = TNODE_P(ttree,i);
	    }
	}
    }
}

/* Setup tree per_level, real_per_level, level_start arrays
   Also pseudo_size, max_one_level, neff_max, one_wide.
*/
static void
setup_levels( tech_tree_t *ttree )
{
    int *per_level, *real_per_level, *level_start;
    tech_node_t *node;
    int i, j, level, prev_level;
    int n_tot, n_real, n_pseudo, n_tot_max;
    float n_eff, n_eff_max;

    per_level = ttree->per_level =
	realloc( ttree->per_level, ttree->n_levels*sizeof(int) );
    real_per_level = ttree->real_per_level =
	realloc( ttree->real_per_level, ttree->n_levels*sizeof(int) );
    level_start = ttree->level_start =
	realloc( ttree->level_start, ttree->n_levels*sizeof(int) );

    CHECK_ALLOC( per_level && real_per_level && level_start );

    for( i=0; i<ttree->n_levels; i++ ) {
	per_level[i] = real_per_level[i] = level_start[i] = 0;
    }

    prev_level = -1;
    for( i=0; i<ttree->n; i++ ) {
	j = ttree->i_sort[i];
	node = ttree->nodes + j;
	level = node->level;
	assert( level>=0 && level<ttree->n_levels
		&& (level==prev_level || level==prev_level+1) );
	if( level>prev_level ) {
	    level_start[level] = i;
	    prev_level = level;
	}
	per_level[level]++;
	if( !node->pseudo ) {
	    real_per_level[level]++;
	}
    }

    ttree->pseudo_size = 0.01*(float)(opt->pseudo_size);
    n_tot_max = 0;
    n_eff_max = 0.0;
    for( i=0; i<ttree->n_levels; i++ ) {
	n_tot = ttree->per_level[i];
	n_real = ttree->real_per_level[i];
	n_pseudo = n_tot - n_real;
	n_eff = n_real + ttree->pseudo_size*n_pseudo;
	if( n_eff > n_eff_max ) {
	    n_eff_max = n_eff;
	}
	if( n_tot > n_tot_max ) {
	    n_tot_max = n_tot;
	}
    }

    ttree->max_one_level = n_tot_max;
    ttree->neff_max = n_eff_max;
    ttree->one_wide = 0.98 / ttree->neff_max;
    /* 0.98 = 1 - (safety margin) */
    if( opt->debug >= DEBUG_DETAIL ) {
	fprintf( debug_out, "neff_max = %g\n" "one_wide = %g\n"
		 "max_one_level = %d\n", ttree->neff_max,
		 ttree->one_wide, ttree->max_one_level );
    }
}

/* Calc node dy values; just depends on level
   Level (n_levels-1) is 0.5/n_levels;
   Level 0 is (n_levels-0.5)/n_levels;
*/
static void
calc_dy( tech_tree_t *ttree )
{
    tech_node_t *node;
    int i;

    for( i=0, node=ttree->nodes; i<ttree->n; i++, node++ ) {
	node->dy = (ttree->n_levels - 0.5 - node->level) / ttree->n_levels;
    }
}

/* a "flexible" (-width) node structure: */
typedef struct {
    int i_tree;			/* node number in ttree */
    float halfwidth;		/* actual halfwidth */
    float z, dz;		/* dx and halfwidth, in progress (for
				   adjusted position and combines */
    int combined;		/* see below */
} flex_node_t;

/* combines:
   We will have a list for the flex_nodes (fnode_tab),
   which are in order from left to right in (initial,final)
   dx. (Where dx is relative position (in [0,1]) on plot.)
   The current dx for each node is z, and the current width
   is dz, which start from dx and halfwidth.
   We adjust positions so the node is within [0,1] (where
   the node has width dz centered at z), and then if any nodes
   are overlapping they get "combined": effectively stuck
   end-to end in order.  This new combined node is then
   fit into place in turn.  When nodes are combined, one
   node get a combined z and dz, and the other gets dz=0
   and z=right-hand-end of combined node.
   The "combined" element of flex_node_t is either: index in
   fnode_tab: fnode is either uncombined or represents a
   combination of nodes (at the "head"); else "combined" is
   less than index in fnode_tab, and in principle gives the
   index of the "head" represents the combination.  Actually
   it may not index the head directly: the node referred to
   may have combined<index in turn, and so on.
*/

/* Calculate dx with more flexible positioning: not all equi-spaced on
   each line if there is enough room and don't "want" to be.
   Input dx value should be (roughly? ideally?) in [0,1].
*/
static void
calc_dx_flexible( tech_tree_t *ttree )
{
    static int first = 1;
    static flex_node_t *fnode_tab;
    static int max_one_level = 0;

    tech_node_t *node;
    flex_node_t *fnode;
    flex_node_t *combn;		/* fnode to combine to */
    int i, j, k;
    int n_this_level, finished;
    float result_dx;
    int iter, iter_bailout = 1000;

    if( first || ttree->max_one_level != max_one_level ) {
	max_one_level = ttree->max_one_level;
	fnode_tab = malloc(max_one_level*sizeof(flex_node_t));
	CHECK_ALLOC( fnode_tab );
	first = 0;
    }
    
    for( i=0; i<ttree->n_levels; i++ ) {
	n_this_level = ttree->per_level[i];
	assert(n_this_level<=max_one_level);
	/* fill in fnode data: */
	for( j=0, fnode=fnode_tab; j<n_this_level; j++, fnode++ ) {
	    k = ttree->i_sort[ttree->level_start[i] + j];
	    node = ttree->nodes + k;
	    fnode->i_tree = k;
	    fnode->halfwidth = 0.5 * TNODE_WIDTH(ttree,node);
	    fnode->z = node->dx;
	    fnode->dz = fnode->halfwidth;
	    fnode->combined = j;
	    if( opt->debug >= DEBUG_DXFLEX ) {
		fprintf( debug_out, "i=%d z=%g, dz=%g\n",
			 fnode->i_tree, fnode->z, fnode->dz );
	    }
	}
	/* combine and adjust until snug: */
	iter = 0;
	do {
	    finished = 1;	/* unless set to 0 below */
	    for( j=0, fnode=fnode_tab; j<n_this_level; j++, fnode++ ) {
		if( fnode->combined != j ) {
		    /* below, combn is either the combined node,
		       or another zero-length node at the end of
		       the combined node. */
		    combn = fnode_tab + fnode->combined;
		    fnode->z = combn->z + combn->dz;
		} else {
		    if( fnode->z - fnode->dz < 0 ) {
			fnode->z = fnode->dz;
			finished = 0;
		    } else if( fnode->z + fnode->dz > 1 ) {
			fnode->z = 1.0 - (1.01*fnode->dz);
			finished = 0;
		    }
		    if( j>0 &&
			fnode->z - fnode->dz < (fnode-1)->z + (fnode-1)->dz ) {
			/* overlap; join them together */
			finished = 0;
			combn = fnode-1;
			while( combn->combined != (combn-fnode_tab) ) {
			    combn = fnode_tab + combn->combined;
			}
			fnode->combined = (combn-fnode_tab);
			combn->z = combn->z * combn->dz +
			    fnode->z * fnode->dz;
			combn->dz += fnode->dz;
			combn->z /= combn->dz;
			fnode->dz = 0;
			fnode->z = combn->z + combn->dz;
		    }
		}
	    }
	    iter++;
	} while( !finished && iter < iter_bailout );
	if( iter >= iter_bailout ) {
	    fprintf( stderr, "WARNING: too many iters in "
		     "calc_dx_flexible()\n" );
	}
	/* uncombine and get the resulting dx values: */
	for( j=0, fnode=fnode_tab; j<n_this_level; j++, fnode++ ) {
	    if( opt->debug >= DEBUG_DXFLEX ) {
		fprintf( debug_out, "o=%d z=%g, dz=%g c=%d\n",
			 fnode->i_tree, fnode->z, fnode->dz, fnode->combined );
	    }
	    if( j==n_this_level-1
		|| (fnode+1)->combined == (j+1) ) {
		result_dx = fnode->z;
	    } else {
		/* take off front, and propogate combined part: */
		result_dx = fnode->z - fnode->dz + fnode->halfwidth;
		(fnode+1)->z = fnode->z + fnode->halfwidth;
		(fnode+1)->dz = fnode->dz - fnode->halfwidth;
	    }
	    ttree->nodes[fnode->i_tree].dx = result_dx;
	    if( opt->debug >= DEBUG_DXFLEX ) {
		fprintf( debug_out, "res=%g\n", result_dx );
	    }
	}
    }
}

/* calculate "preferred" dx based on previous dx values: */
static float
calc_dx_pref( tech_tree_t *ttree, tech_node_t *node )
{
    float dx_sum, weight_sum;
    int k;

    weight_sum = ITER_SELF_WEIGHT;
    dx_sum = node->dx_prev * weight_sum;

    for( k=0; k<node->n_req; k++ ) {
	weight_sum += 1.0;
	dx_sum += TNODE_P(ttree,node->i_req[k])->dx_prev;
    }
    for( k=0; k<node->n_allow; k++ ) {
	weight_sum += 1.0;
	dx_sum += TNODE_P(ttree,node->i_allow[k])->dx_prev;
    }
    return dx_sum / weight_sum;
}

static void
iterate_adjust_dx( tech_tree_t *ttree )
{
    tech_node_t *node;
    int i;

    for( i=0, node=ttree->nodes; i<ttree->n; i++, node++ ) {
	node->dx_prev = node->dx;
    }
    /* use dx so new are based on all old before iteration: */
    for( i=0, node=ttree->nodes; i<ttree->n; i++, node++ ) {
	node->dx = calc_dx_pref( ttree, node );
    }
    for( i=0, node=ttree->nodes; i<ttree->n; i++, node++ ) {
	node->dx = node->dx_prev +
	    ITERSMOOTH_UNDERSHOOT*(node->dx - node->dx_prev);
    }
}

static int
badswap_choice( float p_old, float p_new, float p_scale )
{
    float x, xn, xe, r;

    assert( p_new >= p_old );
    if( BADSWAP_A == 0 ) {
	return 0;
    }
    x = (p_new - p_old) / p_scale;
    if( BADSWAP_N == 1 ) {
	xn = x;
    } else if( BADSWAP_N == 2 ) {
	xn = x*x;
    } else {
	xn = pow( x, BADSWAP_N );
    }
    xe = BADSWAP_A*exp(-badswap_k*xn);
    r = RAND_U01;
    if( opt->debug >= DEBUG_SWAP_EACH ) {
	fprintf( debug_out, "swap p %g %g %g x %g %g %g r %g\n",
		 p_old, p_new, p_scale, x, xn, xe, r );
    }
    return (r < xe);
}

float
dist_nodes_links( tech_tree_t *ttree, tech_node_t *node1,
		  tech_node_t *node2 )
{
    if( SWAP_ON_DIST==2 ) {
	return distx_node_links(ttree, node1)
	    + distx_node_links(ttree, node2);
    } else if( SWAP_ON_DIST ) {
	return dist_node_links(ttree, node1)
	    + dist_node_links(ttree, node2);
    } else {
	return cross_node_links(ttree, node1)
	    + (float)cross_node_links(ttree, node2);
    }
}

/* Swap the dx values for two adjacent nodes at the same level.
   BUT, take care to adjust the dx values to avoid overlaps
   if they are of different "width", so we won't need to
   calc_dx_flexible again after swapping.
*/
static void
swap_node_dx( tech_tree_t *ttree, tech_node_t *node1,
	      tech_node_t *node2 )
{
    float sz1, sz2, x1, x2;

    sz1 = 0.5*TNODE_WIDTH(ttree, node1);
    sz2 = 0.5*TNODE_WIDTH(ttree, node2);
    x1 = node1->dx;
    x2 = node2->dx;
    assert(x1!=x2);
    if( sz1 == sz2 ) {
	node2->dx = x1;
	node1->dx = x2;
    } else {
	if( x1 < x2 ) {
	    node2->dx = x1 - sz1 + sz2;
	    node1->dx = x2 + sz2 - sz1;
	} else {
	    node2->dx = x1 + sz1 - sz2;
	    node1->dx = x2 - sz2 + sz1;
	}
    }
}

/* try swaping adjacent nodes to reduce total path length
   ttree should be sorted by level and dx, and should
   have dy and dx calculated; output changes dx values
   This could be made ~ 25% faster with more bookkeeping?
*/
void
iterate_swaps( tech_tree_t *ttree )
{
    tech_node_t *node1, *node2;
    int i, j, k1, k2, n_this_level, do_swap;
    int i0, j0, rev_ord_levels, rev_ord_this;
    float p_old, p_new;
    int n_good, n_bad, n_bad_do;
    float frac;

    n_good = n_bad = n_bad_do = 0;

    rev_ord_levels = SWAP_RAND_REV_ORDER && (RAND_U01 < 0.5);
    for( i0=0; i0<ttree->n_levels; i0++ ) {
	i = rev_ord_levels ? (ttree->n_levels-1-i0) : i0;
	n_this_level = ttree->per_level[i];
	rev_ord_this = SWAP_RAND_REV_ORDER && (RAND_U01 < 0.5);

	for( j0=0; j0<n_this_level-1; j0++ ) {
	    j = rev_ord_this ? (n_this_level-2-j0) : j0;
	    k1 = ttree->i_sort[ttree->level_start[i]+j];
	    k2 = ttree->i_sort[ttree->level_start[i]+j+1];
	    node1 = TNODE_P(ttree, k1);
	    node2 = TNODE_P(ttree, k2);
	    p_old = dist_nodes_links( ttree, node1, node2 );

	    /* swap and re-calc path length : */
	    swap_node_dx( ttree, node1, node2 );
	    p_new = dist_nodes_links( ttree, node1, node2 );

	    if( p_new < p_old ) {
		n_good++;
		do_swap = 1;
	    } else {
		n_bad++;
		do_swap = badswap_choice( p_old, p_new,
					  (SWAP_ON_DIST?ttree->one_wide:1.0) );
		if( do_swap ) {
		    n_bad_do++;
		}
	    }
	    if( do_swap ) {
		/* already swapped dx;  change i_sort: */
		ttree->i_sort[ttree->level_start[i]+j] = k2;
		ttree->i_sort[ttree->level_start[i]+j+1] = k1;
	    } else {
		/* don't swap: restore dx values: */
		swap_node_dx( ttree, node1, node2 );
	    }
	}
    }
    if( opt->debug >= DEBUG_SWAP_STATS ) {
	fprintf( debug_out, "%d %d %d -1 # good,bad,bad_do\n",
		 n_good, n_bad, n_bad_do );
    }
    if( n_bad > 0 ) {
	frac = n_bad_do/(float)n_bad;
	/* higher k means lower frac */
	if( frac < badswap_f ) {
	    badswap_k /= BADSWAP_K_FACTOR;
	} else {
	    badswap_k *= BADSWAP_K_FACTOR;
	}
	if( BADSWAP_K_MAX && (badswap_k > BADSWAP_K_MAX ) ) {
	    badswap_k = BADSWAP_K_MAX;
	}
	if( opt->debug >= DEBUG_SWAP_STATS2 ) {
	    fprintf( debug_out, "%g %g %g -2 # frac, target, new_k\n",
		     frac, badswap_f, badswap_k );
	}
    }
}

static void
calc_tree_once( tech_tree_t *ttree, int seed )
{
    int i, j;
    
    opt->internal_seed = seed;
    srand( seed );
    
    if( opt->debug >= DEBUG_FINAL_METRICS ) {
	fprintf( debug_out, "# seed = %d\n", seed );
    }

    /* avoid dependence on initial order: (eg from previous seed) */
    sort_techtree( ttree, tcompare_lev_i );
    
    calc_new_layout( ttree, opt, debug_out );
    sort_techtree( ttree, tcompare_lev_dx );
    if( opt->n_iter != 0 ) {
	/* use n_iter=-1 to do this without iters */
	calc_dx_flexible( ttree );
    }

    if( opt->debug >= DEBUG_PROG_METRICS ) {
	calc_print_tree_metrics( debug_out, ttree, 0, "initial" );
    }

    badswap_f = BADSWAP_F_INITIAL;
    badswap_k = BADSWAP_K_INITIAL;
    
    for( i=0; i<opt->n_iter+(opt->n_iter>0)*opt->n_iter_zero; i++ ) {
	if( i>=opt->n_iter ) {
	    badswap_f = 0.0;
	} else if( opt->n_iter == 1 ) {
	    badswap_f = BADSWAP_F_INITIAL;
	} else {
	    badswap_f = BADSWAP_F_INITIAL +
		(BADSWAP_F_FINAL-BADSWAP_F_INITIAL)*i/(float)(opt->n_iter-1);
	}
	for( j=0; j<opt->n_iter_swap; j++ ) {
	    iterate_swaps( ttree );
	    /* iterate_swaps preserves sort, and now also
	       adjusts spacing so new dx_flexible not required */
	    if( opt->debug >= DEBUG_PROG_METRICS ) {
		calc_print_tree_metrics( debug_out, ttree, 1, "swap" );
	    }
	}
	for( j=0; j<opt->n_iter_smooth; j++ ) {
	    iterate_adjust_dx( ttree );
	    if( ITERSMOOTH_REORDER ) {
		sort_techtree( ttree, tcompare_lev_dx );
	    }
	    calc_dx_flexible( ttree );
	    if( opt->debug >= DEBUG_PROG_METRICS ) {
		calc_print_tree_metrics( debug_out, ttree, 2, "itersmooth" );
	    }
	}
    }
    
    calc_tree_metrics( ttree );
    if( opt->debug >= DEBUG_FINAL_METRICS ) {
	print_tree_metrics( debug_out, ttree, 9, "final" );
    }
    if( opt->debug >= DEBUG_DETAIL ) {
	print_tree_sort_dx( debug_out, ttree );
	print_full_tree( debug_out, ttree );
    }
}

/************************* MAIN: ***********************/

int
main( int argc, char *argv[] )
{
    tech_tree_t thetree, *ttree = &thetree;
    float metric, best_metric;
    int i, best_i, *seeds;	/*  */

    get_options( &argc, &argv, opt );
    debug_out = opt->printps ? stderr : stdout;

    if( opt->debug >= DEBUG_OPT_FINAL ) {
	print_options( debug_out, opt, version, "# " );
    }
    
    /* Load rulesets and set some extra strings for some techs: */
    load_rulesets(opt->rulesetdir);
    setup_extras();

    if( opt->rm_redundant ) {
	/* remove redundant reqs: */
	rm_redundant( ttree, (opt->debug>0), debug_out );
    } 	

    /* Convert from advances array to thetree: */
    advances_to_techtree( ttree );

    /* Calculate reverse dependencies: ("allow") */
    calc_allow( ttree );
    
    if( !opt->rm_redundant ) {
	mark_redundant( ttree );
    }

    /* Non-random calcs: */
    calc_levels( ttree );
    add_pseudo_techs( ttree );
    sort_techtree( ttree, tcompare_levels );
    setup_levels( ttree );
    calc_dy( ttree );

    if( opt->n_seed <= 1 ) {
	calc_tree_once( ttree, opt->seed );
    } else {
	srand( opt->seed );
	seeds = malloc(opt->n_seed*sizeof(int));
	CHECK_ALLOC(seeds);
	for( i=0; i<opt->n_seed; i++ ) {
	    seeds[i] = rand();
	}
	best_i = -1;
	best_metric = 0;
	for( i=0; i<opt->n_seed; i++ ) {
	    calc_tree_once( ttree, seeds[i] );
	    metric = (SWAP_ON_DIST==2) ? ttree->metric_distx :
		(SWAP_ON_DIST==1) ? ttree->metric_dist :
		ttree->metric_cross;
	    if( best_i==-1 || metric < best_metric ) {
		best_metric = metric;
		best_i = i;
	    }
	}
	fprintf( debug_out, "best seed=%d\n", seeds[best_i] ); 
	if( best_i != opt->n_seed-1 ) {
	    calc_tree_once( ttree, seeds[best_i] );
	}
    }
    if( opt->printps ) {
	print_ps( ttree, opt );
    }

    return 0;

}

